#include "config.h"
icon_bochs ICON build/win32/nsis/bochs.ico
#include "wx/msw/wx.rc"
#if BX_WITH_WIN32
#include "win32res.rc"
#else
#if BX_DEBUGGER_GUI
#include "win32_enh_dbg.rc"
#endif
#include "bxversion.rc"
#endif
