#define ICON_BOCHS  1000
#define ICON_BOCHS2 1001

#define MAINMENU_DLG 1100
#define IDCONFIG     1110
#define IDREADRC     1120
#define IDWRITERC    1130
#define IDEDITCFG    1140
#define IDRESETCFG   1150
#define IDEDITGRP    1160
#define IDEDITBOX    1170
#define IDSIMU       1180
#define IDRESTORE    1190
#define IDQUIT       1195
#define LOGOPT_DLG   1200
#define IDDEVLIST    1210
#define IDLOGLBL1    1221
#define IDLOGLBL2    1222
#define IDLOGLBL3    1223
#define IDLOGLBL4    1224
#define IDLOGLBL5    1225
#define IDLOGEVT1    1231
#define IDLOGEVT2    1232
#define IDLOGEVT3    1233
#define IDLOGEVT4    1234
#define IDLOGEVT5    1235
#define IDADVLOGOPT  1240
#define IDAPPLY      1250
#define PLUGIN_CTRL_DLG 1300
#define IDPLUGTXT1   1305
#define IDPLUGLIST1  1310
#define IDPLUGTXT2   1315
#define IDPLUGLIST2  1320
#define IDLOAD       1330
#define IDUNLOAD     1340
#define ASK_DLG     2000
#define IDASKTX1    2010
#define IDASKDEV    2020
#define IDASKTX2    2030
#define IDASKMSG    2040
#define IDASKLIST   2050
#define STRING_DLG  2100
#define IDSTRING    2110
#define PARAM_DLG   3000
#define IDCREATE    3010
