#define ISOLATION_AWARE_ENABLED
#include <windows.h>
#include "config.h"
#include "gui/win32res.h"

ICON_BOCHS ICON build/win32/nsis/bochs.ico
ICON_BOCHS2 ICON build/win32/nsis/logo.ico

// Manifest for both 32-bit and 64-bit Windows
#if !defined(VSIDE) && !BX_WITH_WX
1 24 build/win32/bochs.manifest
#endif

ASK_DLG DIALOG 30, 30, 250, 100
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Event"
FONT 8, "Helv"
BEGIN
    LTEXT           "Device", IDASKTX1, 10, 12, 40, 14
    EDITTEXT        IDASKDEV, 45, 10, 195, 14, ES_READONLY
    LTEXT           "Message", IDASKTX2, 10, 27, 40, 14
    EDITTEXT        IDASKMSG, 45, 25, 195, 14, ES_READONLY | ES_AUTOHSCROLL
    LISTBOX         IDASKLIST, 35, 50, 120, 45, WS_VSCROLL | WS_TABSTOP
    DEFPUSHBUTTON   "OK", IDOK, 165, 50, 50, 14
    PUSHBUTTON      "Cancel", IDCANCEL, 165, 70, 50, 14
END

STRING_DLG DIALOG 30, 30, 130, 65
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Param"
FONT 8, "Helv"
BEGIN
    EDITTEXT        IDSTRING, 15, 15, 100, 14
    DEFPUSHBUTTON   "OK", IDOK, 10, 40, 50, 14
    PUSHBUTTON      "Cancel", IDCANCEL, 70, 40, 50, 14
END

MAINMENU_DLG DIALOG 30, 30, 275, 130
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Bochs Start Menu"
FONT 8, "Helv"
BEGIN
    GROUPBOX        "Configuration", IDCONFIG, 10, 10, 70, 97
    PUSHBUTTON      "L&oad", IDREADRC, 20, 25, 50, 14
    PUSHBUTTON      "&Save", IDWRITERC, 20, 45, 50, 14
    PUSHBUTTON      "&Edit", IDEDITCFG, 20, 65, 50, 14
    PUSHBUTTON      "Rese&t", IDRESETCFG, 20, 85, 50, 14
    GROUPBOX        "Edit Options", IDEDITGRP, 90, 10, 95, 107
    LISTBOX         IDEDITBOX, 95, 20, 85, 100, WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Simulation", IDSIMU, 195, 10, 70, 77
    DEFPUSHBUTTON   "Sta&rt", IDOK, 205, 25, 50, 14
    PUSHBUTTON      "Restore St&ate", IDRESTORE, 205, 45, 50, 14
    PUSHBUTTON      "&Quit", IDQUIT, 205, 65, 50, 14
END

LOGOPT_DLG DIALOG 30, 30, 240, 180
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Log Options"
FONT 8, "Helv"
BEGIN
    LISTBOX         IDDEVLIST, 15, 25, 55, 100, LBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Debug events", IDLOGLBL1, 80, 27, 55, 14
    COMBOBOX        IDLOGEVT1, 155, 25, 65, 56, CBS_DROPDOWNLIST | CBS_AUTOHSCROLL |
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Info events", IDLOGLBL2, 80, 47, 55, 14
    COMBOBOX        IDLOGEVT2, 155, 45, 65, 56, CBS_DROPDOWNLIST | CBS_AUTOHSCROLL |
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Error events", IDLOGLBL3, 80, 67, 55, 14
    COMBOBOX        IDLOGEVT3, 155, 65, 65, 56, CBS_DROPDOWNLIST | CBS_AUTOHSCROLL |
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Panic events", IDLOGLBL4, 80, 87, 55, 14
    COMBOBOX        IDLOGEVT4, 155, 85, 65, 56, CBS_DROPDOWNLIST | CBS_AUTOHSCROLL |
                    WS_VSCROLL | WS_TABSTOP
    AUTOCHECKBOX    "Specify log options per device", IDADVLOGOPT, 50, 135, 112, 14, BS_LEFTTEXT | WS_TABSTOP
    DEFPUSHBUTTON   "OK", IDOK, 35, 155, 50, 14
    PUSHBUTTON      "Cancel", IDCANCEL, 95, 155, 50, 14
    PUSHBUTTON      "Apply", IDAPPLY, 155, 155, 50, 14
END

PLUGIN_CTRL_DLG DIALOG 100, 120, 280, 140
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Optional Plugin Control"
FONT 8, "Helv"
BEGIN
    LTEXT           "Available", IDPLUGTXT1, 40, 10, 55, 14
    LISTBOX         IDPLUGLIST1, 15, 20, 85, 100, WS_VSCROLL | WS_TABSTOP
    LTEXT           "Loaded", IDPLUGTXT2, 200, 10, 55, 14
    LISTBOX         IDPLUGLIST2, 175, 20, 85, 100, WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      ">> Load >>", IDLOAD, 110, 40, 50, 14
    PUSHBUTTON      "<< Unload <<", IDUNLOAD, 110, 60, 50, 14
    DEFPUSHBUTTON   "OK", IDOK, 115, 120, 50, 14
END

PARAM_DLG DIALOG 30, 30, 200, 65
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Param"
FONT 8, "Helv"
BEGIN
    DEFPUSHBUTTON   "OK", IDOK, 45, 40, 50, 14
    PUSHBUTTON      "Cancel", IDCANCEL, 105, 40, 50, 14
END

#include "bxversion.rc"
#if BX_DEBUGGER_GUI
#include "win32_enh_dbg.rc"
#endif
